/*
 * Copyright (c) 2016 Intel Corporation
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/drivers/spi.h>

#define SPI_FREQUENCY		1000000
#define SPI_TRANSACTION_LENGTH	7
#define INDICATE_ALIVE_FREUENCY 100000

#define LED0_NODE DT_ALIAS(led0)
#define LED1_NODE DT_ALIAS(led1)

static const struct gpio_dt_spec alive_led = GPIO_DT_SPEC_GET(LED0_NODE, gpios);
static const struct gpio_dt_spec error_led = GPIO_DT_SPEC_GET(LED1_NODE, gpios);

#define MY_SPI_MASTER DT_NODELABEL(my_spi_master)
#define MY_SPI_MASTER_CS_DT_SPEC SPI_CS_GPIOS_DT_SPEC_GET(DT_NODELABEL(reg_my_spi_master))

#define MY_SPI_SLAVE  DT_NODELABEL(my_spi_slave)

static uint8_t master_tx_buffer[SPI_TRANSACTION_LENGTH];
static uint8_t slave_rx_buffer[SPI_TRANSACTION_LENGTH];

// SPI master functionality
const struct device *spi_dev;
static struct k_poll_signal spi_done_sig = K_POLL_SIGNAL_INITIALIZER(spi_done_sig);

static void print_array(const uint8_t *in_data, uint32_t in_len)
{
	printk(" (size %lu):", (unsigned long)in_len);
	if (NULL != in_data) {
		for (uint32_t i = 0; i < in_len; i++) {
			if (i % 16 == 0)
				printk("\n\t%02X ", in_data[i]);
			else
				printk("%02X ", in_data[i]);
		}
	}
	printk("\n");
}

static void spi_init(void)
{
	spi_dev = DEVICE_DT_GET(MY_SPI_MASTER);
	if(!device_is_ready(spi_dev)) {
		printk("SPI master device not ready!\n");
	}
	struct gpio_dt_spec spim_cs_gpio = MY_SPI_MASTER_CS_DT_SPEC;
	if(!device_is_ready(spim_cs_gpio.port)){
		printk("SPI master chip select device not ready!\n");
	}
}

static struct spi_config spi_cfg = {
	.operation = SPI_WORD_SET(8) | SPI_TRANSFER_MSB | SPI_MODE_CPOL | SPI_MODE_CPHA,
	.frequency = SPI_FREQUENCY,
	.slave = 0,
	.cs = {.gpio = MY_SPI_MASTER_CS_DT_SPEC, .delay = 0},
};

static int spi_write_test_msg(uint32_t counter)
{
	const struct spi_buf tx_buf = {
		.buf = master_tx_buffer,
		.len = sizeof(master_tx_buffer)
	};
	const struct spi_buf_set tx = {
		.buffers = &tx_buf,
		.count = 1
	};

	// Update the TX buffer with a rolling counter
	master_tx_buffer[0] = counter;
	master_tx_buffer[1] = (counter >> 8) & 0xFF;
	master_tx_buffer[2] = (counter >> 16) & 0xFF;
	master_tx_buffer[3] = (counter >> 24) & 0xFF;

	// Fill the next 3 words with somethign other than 0...
	master_tx_buffer[4] = 0xDE;
	master_tx_buffer[5] = 0xCA;
	master_tx_buffer[6] = 0xDE;

	if (counter % INDICATE_ALIVE_FREUENCY == 0) {
		printk("Starting transaction %u...\n", counter);
	}

	// Reset signal
	k_poll_signal_reset(&spi_done_sig);
	
	// Start transaction
	int error = spi_write_signal(spi_dev, &spi_cfg, &tx, &spi_done_sig);
	if (error != 0){
		printk("SPI transceive error: %i\n", error);
		return error;
	}

	// Wait for the done signal to be raised
	int spi_signaled, spi_result;
	do {
		k_poll_signal_check(&spi_done_sig, &spi_signaled, &spi_result);
	} while(spi_signaled == 0);

	return 0;
}

// SPI slave functionality
const struct device *spi_slave_dev;
static struct k_poll_signal spi_slave_done_sig = K_POLL_SIGNAL_INITIALIZER(spi_slave_done_sig);

static const struct spi_config spi_slave_cfg = {
	.operation = SPI_WORD_SET(8) | SPI_TRANSFER_MSB |
				 SPI_MODE_CPOL | SPI_MODE_CPHA | SPI_OP_MODE_SLAVE,
	.frequency = SPI_FREQUENCY,
	.slave = 0,
};

static void spi_slave_init(void)
{
	spi_slave_dev = DEVICE_DT_GET(MY_SPI_SLAVE);
	if (!device_is_ready(spi_dev)) {
		printk("SPI slave device not ready!\n");
	}
}

static int spi_slave_read_prepare(void)
{
	struct spi_buf s_rx_buf = {
		.buf = slave_rx_buffer,
		.len = sizeof(slave_rx_buffer),
	};
	const struct spi_buf_set s_rx = {
		.buffers = &s_rx_buf,
		.count = 1
	};

	// Reset signal
	k_poll_signal_reset(&spi_slave_done_sig);
	
	// Start transaction
	int error = spi_read_signal(spi_slave_dev, &spi_slave_cfg, &s_rx, &spi_slave_done_sig);
	if (error != 0){
		printk("SPI slave error: %i\n", error);
		return error;
	}
	return 0;
}

static int spi_slave_check_for_message(void)
{
	int signaled, result;
	k_poll_signal_check(&spi_slave_done_sig, &signaled, &result);
	if (signaled != 0){
		return result; // Received byte count
	}
	else return -1;
}

int main(void)
{
	int ret;
	int spis_received_count;
	uint32_t transaction_counter = 0;

	if (!device_is_ready(alive_led.port)) {
		return 0;
	}

	ret = gpio_pin_configure_dt(&alive_led, GPIO_OUTPUT_ACTIVE);
	if (ret < 0) {
		return 0;
	}

	if (!device_is_ready(error_led.port)) {
		return 0;
	}

	ret = gpio_pin_configure_dt(&error_led, GPIO_OUTPUT_ACTIVE);
	if (ret < 0) {
		return 0;
	}

	gpio_pin_set_dt(&error_led, 0);
	spi_init();
	spi_slave_init();
	printk("SPI master/slave example started\n");
	spi_slave_read_prepare();

	while (1) {
		spi_write_test_msg(transaction_counter);
		gpio_pin_toggle_dt(&alive_led);

		spis_received_count = spi_slave_check_for_message();

		if (spis_received_count == SPI_TRANSACTION_LENGTH &&
		    memcmp(slave_rx_buffer, master_tx_buffer, SPI_TRANSACTION_LENGTH) == 0) {
			// All good. Prepare the next SPI slave transaction
			transaction_counter++;
			spi_slave_read_prepare();
		} else {
			// Indicate error (can be used to trigger logic analyzer)
			gpio_pin_set_dt(&error_led, 1);

			printk("Error during transaction %u!\n", transaction_counter);
			printk("Expected");
			print_array(master_tx_buffer, SPI_TRANSACTION_LENGTH);
			printk("Received");
			print_array(slave_rx_buffer, spis_received_count);

			break;
		}
	}

	return 0;
}
